/**************************************************************************
* The following is to enable dumps from inside JS.  If I can get this
* working I can use this to debug the code.
**************************************************************************/
var debugboolean = true;
//setting pref for debug
if(debugboolean==true)
{
    const PREFS_CID      = "@mozilla.org/preferences;1";
    const PREFS_I_PREF   = "nsIPref";
    const PREF_STRING    = "browser.dom.window.dump.enabled";
    try {
	var Pref        = new Components.Constructor(PREFS_CID, PREFS_I_PREF);
	var pref        = new Pref( );
	pref.SetBoolPref(PREF_STRING, true);
    } catch(e) { 
	alert(e);
    }
}

// Some XPCOM constants we need to walk the bookmarks tree
// Commented out because they are already defined in chrome://communicator/content/bookmarks/bookmarks.js
// const RDF = Components.classes['@mozilla.org/rdf/rdf-service;1'].getService(Components.interfaces.nsIRDFService);
// const RDFC = Components.classes['@mozilla.org/rdf/container;1'].createInstance(Components.interfaces.nsIRDFContainer);
// const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";
// const NC_NS = "http://home.netscape.com/NC-rdf#";
// const BMDS = Components .classes["@mozilla.org/rdf/datasource;1?name=bookmarks"].getService(Components.interfaces.nsIRDFDataSource); 
// const RDFCU = Components.classes['@mozilla.org/rdf/container-utils;1'].getService(Components.interfaces.nsIRDFContainerUtils);
// const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";
// const NC_NS = "http://home.netscape.com/NC-rdf#";
// const RDF_NS = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
const XPPREF = Components.classes["@mozilla.org/preferences;1"].getService(Components.interfaces.nsIPrefBranch);
const USER_AGENT = navigator.appCodeName + "/" + navigator.appVersion + " Habari Xenu News Aggregator";


// Some global variables that are used across functions
var bookmarksTree;	// The tree used to display the feeds
var xmlhttp;		// A HTTP request object
var localPath;		// Path to feeds downloaded into a local file
var win;		// The Habari Xenu window

// HTTP request results
const RESULT_OK = 0;
const RESULT_PARSE_ERROR = 1;
const RESULT_NOT_RSS = 2;
const RESULT_NOT_FOUND = 3;
const RESULT_NOT_AVAILABLE = 4;
const RESULT_ERROR_FAILURE = 5;



/**************************************************************************
* This function gets called when we need to open the settings editor
**************************************************************************/
function openSettingsEditor() {
    dump("openSettingsditor(): Opening Settings editor\n");

    var settingsURL = "chrome://habarixenu/content/settingseditor.xul";

    // Open a new modal dialog showing the settingseditor xul as chrome
    // pass it a reference to the bookmarksTree
    window.openDialog(settingsURL,"","chrome,modal", bookmarksTree);

    dump("openSettingsditor(): Finished opening Settings editor\n");
    return true;
}



/**************************************************************************
* Close the application
**************************************************************************/
function closeHabariXenu() {
    dump("closeHabariXenu(): Closing window\n");
    window.close();
}



/**************************************************************************
* Populate the tree from the Datasource. See above for how we get 'ds'
* This function gets called on the XUL window's onload event
**************************************************************************/
function buildTreeFromDataSource() {
    dump("buildTreeFromDataSource(): Beginning to build subscription tree from bookmarks\n");

    // Initialize win to point to the main window
    win = document.getElementsByTagName("window")[0];

    // None of the following seem to work correctly
    // win.setAttribute("wait-cursor", "true");
    // win.setAttribute("wait-cursor", "false");
    // this.setCursor("wait");

    // Activate the progress meter to indicate work happening
    var progmtr = document.getElementById("progmtr");
    progmtr.setAttribute("mode", "undetermined");

    // Initialize some stuff in chrome://communicator/content/bookmarks/bookmarks.js
    initServices();
    initBMService();

    // Set the bookmarksTree to point to the tree in which feeds are stored
    bookmarksTree = document.getElementById("bookmarksTree");

    var prefFolder;

    // Has user selected one folder as preferred feeds folder
    if(XPPREF.getPrefType("habarixenu.feeds_folder") == XPPREF.PREF_INVALID){
	// No preferred folder.  Set it to the top of bookmarks RDF
	dump("buildTreeFromDataSource(): No preference.  Setting to top of bookmarks tree.\n");
	prefFolder = "NC:BookmarksRoot";
    } else {
	// Set to the preferred folder
	prefFolder = XPPREF.getCharPref("habarixenu.feeds_folder").toString();
	dump("buildTreeFromDataSource(): Existing preference.  Setting to " + prefFolder + "\n");
    }

    // Set the attribute on the tree and rebuild it
    bookmarksTree.setAttribute("ref", prefFolder);
    bookmarksTree.builder.rebuild();

    // Deactivate the progress meter to indicate work done
    progmtr.setAttribute("mode", "determined");
    progmtr.setAttribute("value", "0%");
    // win.setAttribute("wait-cursor", "false");
    // win.setAttribute("default-cursor", "true");
    // win.setCursor("default");

    dump("buildTreeFromDataSource(): Finished rebuilding subscription tree from bookmarks.\n");
}



/**************************************************************************
* This function gets called when the user selects an item in the feeds tree
* Here we find the URL and last update time of the feed.
**************************************************************************/
function fetchFeed(bmEvent) {
    dump("fetchFeed(): Entered.\n");

    // A flag to indicate whether we downloaded a fresh local copy of a remote feed
    var freshDownload = false;

    // If nothing was selected we have nothing to do
    if (bookmarksTree.currentIndex < 0) {
	dump("fetchFeed(): Nothing selected.  Returning.\n");
	return;
    }

    // If folder was selected we have nothing to do
    var fetchURL = bookmarksTree.view.getCellText(bookmarksTree.currentIndex, "feedlist-url");
    if (! fetchURL) {
	dump("fetchFeed(): Folder selected.  Returning.\n");
	return false;
    }


    // Activate the progress meter to indicate work happening
    var progmtr = document.getElementById("progmtr");
    progmtr.setAttribute("mode", "undetermined");
    progmtr.setAttribute("value", "50%");

    // Switch to busy/wait cursor
    // win.setAttribute("wait-cursor", "true");


    // The URN of the bookmark item that has been selected
    var selectURN = bookmarksTree.builderView.getResourceAtIndex(bookmarksTree.currentIndex).Value;

    const rName = RDF.GetResource(WEB_NS + "LastVisitDate");
    var rSource = RDF.GetResource(selectURN);
    var node =  BMDS.GetTarget(rSource, rName, true);

    var lastVisited;
    try {
	// Get the LastVisitDate for this bookmark if it exists
	lastVisited = node.QueryInterface(Components.interfaces.nsIRDFDate).Value;
	lastVisited /= 1000;	// nsIRDFDate is in microseconds JS Date() deals with milliseconds
    } catch (e) {
	// LastVisitDate does not exist for this bookmark.  Probably a newly 
	// added bookmark?  Assume it was created just after the moment of creation
	lastVisited = new Date(1);  // One second since epoch
    }

    var lastUpdatedAt = new Date(lastVisited);
    dump("fetchFeed(): Feed was last updated at " + lastUpdatedAt + "\n");

    var currentTime = new Date();
    dump("fetchFeed(): Current time is " + currentTime + "\n");

    // TODO We must not hard code 1 hour.  It must be user-settable
    // 1000 milliseconds in every second
    // 60 seconds in every minute
    // 60 minutes in every hour
    const oneHour = 1000 * 60 * 60 * 1;

    var interval = currentTime - lastUpdatedAt;
    dump("fetchFeed(): interval since last update is " + interval + " ms\n");

    localPath = absPathToLocal(fetchURL);

    // Has it been an hour since the last update?
    if (interval > oneHour) {
	dump("fetchFeed(): Feed is stale.  Needs update.\n");
	downloadFromURL(fetchURL, localPath);
	freshDownload = true;
    } else {
	dump("fetchFeed(): Feed is supposed to be fresh.  Let us see if we have a local copy.\n");

	// Do we need to download a fresh copy from the server?
	try {
	    var localFile = new File(localPath);
	    if (localFile.exists()) {
		dump("fetchFeed(): We still have a local copy of the feed.  Going to parse it.\n")
		parseFeed();
	    } else {
		dump("fetchFeed(): We don't have a local copy of the feed.  Will fetch again.\n")
		downloadFromURL(fetchURL, localPath);
		freshDownload = true;
	    }
	} catch (e) {
	    dump("fetchFeed(): Error occured trying to see if local feed file " + localPath + " exists.\n");
	    dump(e + "\n");
	}
    }

    if (freshDownload) {
	// Newly downloaded feed.  Set its LastVisitDate in the BMDS
	dump("fetchFeed(): This is a newly downloaded feed.  Setting LastVisitDate in bookmarks.\n");

	var currentmicroseconds = currentTime.valueOf() * 1000;
	var newNode = RDF.GetDateLiteral(currentmicroseconds);

	if (node) {
	    // A LastVisitDate exists for this bookmark.  Change it
	    try {
		dump("Called with " + rSource + "," + rName + "," + node + "," + newNode + "\n");
		BMDS.Change(rSource, rName, node, newNode);
		BMDS.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
	    } catch (e) {
		dump("fetchFeed(): Error changing LastVisitDate of bookmark: " + e + "\n");
	    }
	} else {
	    // A LastVisitDate does not exist for this bookmark.  Create it
	    try {
		dump("Called with " + rSource + "," + rName + "," + newNode + "\n");
		BMDS.Assert(rSource, rName, newNode, true);
		BMDS.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
	    } catch (e) {
		dump("fetchFeed(): Error changing LastVisitDate of bookmark: " + e + "\n");
	    }
	}

    }

    // Reset feed
    feed = new Object;

    // Deactivate the progress meter to indicate work done
    progmtr.setAttribute("mode", "determined");
    progmtr.setAttribute("value", "0%");

    dump("fetchFeed(): Finished. Exiting.\n");
    return false;   // Returning false because we want this event to bubble up
}



/**************************************************************************
* A utility function that takes a URl, converts into an md5 hash, appends
* it to a local feeds directory location and returns the resulting string
**************************************************************************/
function absPathToLocal(url) {
    dump("absPathToLocal(): Called with " + url + "\n");
    var fu = new FileUtils();
    var hash = hex_md5(url);

    var chromeFolder = fu.chromeToPath("chrome://habarixenu/content/feeds/");
    var fullPath = "file:" + chromeFolder + hash;

    dump("absPathToLocal(): Returning " + fullPath + "\n");
    return fullPath;
}



/**************************************************************************
* A function to download url into localPath.  It uses XPCOM.  We currently
* do not use it.  It is here purely for reference purposes.  We might 
* decide to use it in the future
**************************************************************************/
function downloadFile(url, localPath) {
     //new obj_URI object
     try {
	 var obj_URI =
	 Components.classes["@mozilla.org/network/standard-url;1"].createInstance(Components.interfaces.nsIURI);
     } catch (e) { alert(e); }

     //set obj_URI
     obj_URI.spec = url;
     //new file object
     try {
	 var obj_TargetFile =
	 Components.classes["@mozilla.org/file/local;1"].createInstance(Components.interfaces.nsILocalFile);
     } catch (e) { alert(e); }

     //set file with path
     obj_TargetFile.initWithPath(localPath);
     //if file doesn't exist, create
     if(!obj_TargetFile.exists)
     {
	 try {
	     obj_TargetFile.create(0x00,0644);
	 } catch (e) { alert(e); }
     }
     //new persitence object
     try {
	 obj_Persist =
	 Components.classes["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"].createInstance(Components.interfaces.nsIWebBrowserPersist);
     } catch (e) { alert(e);}
     //new doenload object & init
     var obj_Download =
     Components.classes["@mozilla.org/progressdialog;1"].createInstance(Components.interfaces.nsIProgressDialog);
     obj_Download.init(obj_URI,obj_TargetFile, null, null, null, obj_Persist);

     //save file to target
     obj_Persist.saveURI(obj_URI,null,null,null,null,obj_TargetFile);
}



/**************************************************************************
* A utility function to download the latest feed from a URL and store it
* locally into localPath
**************************************************************************/
function downloadFromURL(url, localPath) {
    dump("downloadFromURL(): Called with: " + url + "," + localPath + "\n");

    xmlhttp = new XMLHttpRequest();

    // On the following events run the named functions
    xmlhttp.onload = xmlhttpLoaded;
    xmlhttp.onerror = xmlhttpError;
    xmlhttp.onreadystatechange = xmlhttpReadyStateChange;

    // Initiate the request
    try {
	xmlhttp.open("GET", url, true);	// NOTE: Calling synchronously
	xmlhttp.setRequestHeader("User-Agent", USER_AGENT);
	xmlhttp.overrideMimeType("text/xml");
    } catch (e) {
	dump(e);
	xmlhttpGetResult(RESULT_ERROR_FAILURE);
    }

    try {
	xmlhttp.send(null);
    } catch (e) {
	dump(e);
	xmlhttpGetResult(RESULT_ERROR_FAILURE);
    }
}

function xmlhttpError(e) {
    alert("HTTP Error " + e);
}

function xmlhttpReadyStateChange() {
    if (xmlhttp.readyState == 2) {
	try {
	    if (xmlhttp.status == 404) {
		xmlhttpGetResult(RESULT_NOT_FOUND);
	    }
	} catch (e) {
	    xmlhttpGetResult(RESULT_NOT_AVAILABLE);
	    return;
	}
    } else if (xmlhttp.readyState == 3) {}
}

function xmlhttpLoaded(e) {
    var localFile;
    try {
	localFile = new File(localPath);
    } catch (e) {
	dump("xmlhttpLoaded(): Error initializing file: " + e + "\n");
    }

    try {
	if (! localFile.exists()) {
	    dump("xmlhttpLoaded(): " + localPath + " does not exist. Creating\n");
	    localFile.create();
	}
    } catch (e) {
	dump("xmlhttpLoaded(): Error creating file " + localPath + " : " + e + "\n");
    }

    if(localFile.open('w')) { 
	dump("downloadFromURL(): Writing fetched data to " + localPath + "\n");
	localFile.write("" + xmlhttp.responseText);
    }
    localFile.close();
    xmlhttpGetResult(RESULT_OK);
}

function xmlhttpGetResult(aResultCode) {
    xmlhttp.abort();
    if (aResultCode == RESULT_OK) {
	// Everything OK.  Will do something here
	parseFeed();
    }
}



/**************************************************************************
* A function that takes the RSS that has been loaded into the hidden
* iframe, goes through it using the DOM and creates an HTML representation
* of the data in the RSS
**************************************************************************/
function parseFeed() {
    // Let us sleep for a second to make sure that the DOM rebuilds itself
    /*
    for (var i = 0; i < 10; i++) {
	setTimeout('', 500);
    }
    */

    var feed = '';
    try {
	dump("parseFeed(): parsing " + localPath + "\n");
	feed=new FeedParser(localPath);
    } catch (e) {
	dump("parseFeed(): Error parsing feed: " + e + "\n");
    }

    // setTimeout('',500);
    var parsedContent = '';

    if (feed.typeRDF) {
	// RDF/RSS 1.0 Feeds
	dump("parseFeed(): feed is RDF\n");

	parsedContent = "<HTML><HEAD>\n";
	var titleText = feed.channel.title;
	var linkText = feed.channel.link;
	var descText = feed.channel.description;
	dump(titleText + "\n");

	parsedContent += "<TITLE>News from " + titleText + "</TITLE>\n";
	parsedContent += "<link rel='stylesheet' type='text/css' href='chrome://habarixenu/skin/display.css'>\n";
	parsedContent += "</HEAD><BODY>\n";

	// TODO must figure out images
	parsedContent += "<div class='header'>\n";
	parsedContent += "<b class='header-title' id='header-title'><a href='" + linkText + "' target='_new'>" + titleText + "</a></b><br>\n";
	parsedContent += descText;
	parsedContent += "</div>\n";
	dump("parseFeed(): title: " + titleText + "\n");
	dump("parseFeed(): description: " + descText + "\n");

	for (var i = 0; i < feed.item.length; i++) {
	    parsedContent += "<div class='item'>\n";
	    if (feed.item[i].title && feed.item[i].link) {
		parsedContent += "<b><a href='" + feed.item[i].link + "' target='_new'>" + feed.item[i].title + "</a></b><br>\n";
	    } else if (feed.item_rdfabout[i] && feed.item[i].title) {
		parsedContent += "<b><a href='" + feed.item_rdfabout[i] + "' target='_new'>" + feed.item[i].title + "</a></b><br>\n";
	    } else {
		parsedContent += "<b><a href='" + feed.item[i].link + "' target='_new'>" + feed.item[i].link + "</a></b><br>\n";
	    }
	    parsedContent += feed.item[i].description;
	    parsedContent += "</div>\n";
	}
    } else if (feed.typeRSS) {
	// RDF/RSS 1.0 Feeds
	dump("parseFeed(): feed is RSS\n");

	parsedContent = "<HTML><HEAD>\n";
	var titleText = feed.channel.title;
	var linkText = feed.channel.link;
	var descText = feed.channel.description;
	dump(titleText + "\n");

	parsedContent += "<TITLE>News from " + titleText + "</TITLE>\n";
	parsedContent += "<link rel='stylesheet' type='text/css' href='chrome://habarixenu/skin/display.css'>\n";
	parsedContent += "</HEAD><BODY onload='return true;'>\n";

	// TODO must figure out images
	parsedContent += "<div class='header'>\n";
	parsedContent += "<b class='header-title' id='header-title'><a href='" + linkText + "' target='_new'>" + titleText + "</a></b><br>\n";
	parsedContent += descText;
	parsedContent += "</div>\n";
	dump("parseFeed(): title: " + titleText + "\n");
	dump("parseFeed(): description: " + descText + "\n");
	dump("parseFeed(): item: " + feed.channel.item.length + "\n");

	for (var i = 0; i < feed.channel.item.length; i++) {
	    parsedContent += "<div class='item'>\n";

	    parsedContent += "<b><a href='" + feed.channel.item[i].link + "' target='_new'>";
	    if (feed.channel.item[i].title) {
		parsedContent += feed.channel.item[i].title + "</a><br>\n";
	    } else {
		parsedContent += feed.channel.item[i].link + "</a><br>\n";
	    }

	    if (feed.channel.item[i].description) {
		parsedContent += feed.channel.item[i].description;
	    }
	    parsedContent += "</div>\n";
	}
    }

    // Open up the visible ibrowser and write the parsed content onto it
    var parsedDocument = window.frames[0].document;
    parsedDocument.body.innerHTML = parsedContent;

    // Switch to normal cursor
    // win.setAttribute("wait-cursor", "false");
    return true;
}



/**************************************************************************
* Show information about Habari Xenu
**************************************************************************/
function loadAbout() {
    var browser =  document.getElementById("parsed-rss-feed");
    browser.loadURI("chrome://habarixenu/locale/about.html");

    return true;
}
